# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Formatting module for outputting captures.
"""

import json
from typing import Iterable, AbstractSet

from .model import Formatter, RowBuilder
from ...damlast.daml_types import con
from ...damlast.protocols import SymbolLookup
from ...prim import Party
from ...values import JsonEncoder, Context
from ...util.tools import boundary_iter


class JsonFormatter(Formatter):

    def render(
            self,
            lookup: 'SymbolLookup',
            parties: 'AbstractSet[Party]',
            entries: 'Iterable[RowBuilder]'):
        """
        Return a list of strings that, when taken together, constitute a JSON document that lists
        all of the entries of the ledger.

        :param lookup:
            The contents to render.
        :param parties:
            The list of of parties.
        :param entries:
            The entries to render.
        """
        context = Context(JsonEncoder(), lookup)

        party_list = sorted(parties)

        yield "{{ \"parties\": {},".format(json.dumps(party_list))
        yield "  \"contracts\": ["

        for is_last, entry in boundary_iter(entries):
            cid = entry.cid
            yield f'  {{ "contract_id": "{cid.value}",'
            yield f'    "template": {json.dumps(str(cid.value_type))},'
            yield f'    "parties": {json.dumps(party_list)},'
            yield f'    "arguments": {json.dumps(context.convert(con(cid.value_type), entry.cdata))} }}' + \
                  ("" if is_last else ",")

        yield "  ] }"



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional, Type, TypeVar

# noinspection PyPackageRequirements
from google.protobuf.empty_pb2 import Empty
# noinspection PyPackageRequirements
from google.protobuf import timestamp_pb2

from .context import Context
from .mapper import ValueMapper
from .._gen.com.daml.ledger.api.v1 import value_pb2
from ..damlast.daml_lf_1 import Type, DefDataType
from ..damlast.util import find_variant_type
from ..prim import decimal_to_str, to_bool, to_date, to_datetime, to_decimal, to_int, to_str, \
    to_variant, date_to_int, datetime_to_epoch_microseconds, to_party

__all__ = ['ProtobufDecoder', 'ProtobufEncoder', 'get_value', 'set_value']

T = TypeVar('T')


class ProtobufDecoder(ValueMapper):
    """
    Convert value_pb2.Value (or any of its cases) to a native Python type.

    This mapper also handles:
     * re-typing ContractIds (they come over the Ledger API without their type parameter)
     * non-verbose Ledger API record values (field names are omitted for message brevity, meaning
       they can only be understood in a generic way by knowing the metadata associated with the
       object ahead of time)
    """

    def data_record(self, context: 'Context', dt: 'DefDataType', record: 'DefDataType.Fields', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'record', value_pb2.Record)

        d = dict()
        for fld_metadata, fld_data in zip(record.fields, msg.fields):
            name = fld_metadata.field
            value = context.append_path(name).convert(fld_metadata.type, fld_data.value)
            d[name] = value

        return d

    def data_variant(self, context: 'Context', dt: 'DefDataType', variant: 'DefDataType.Fields', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'variant', value_pb2.Variant)
        obj_ctor = msg.constructor
        obj_value = msg.value
        obj_type = find_variant_type(dt, variant, obj_ctor)

        d_value = context.append_path(obj_ctor).convert(obj_type, obj_value)

        return {obj_ctor: d_value}

    def data_enum(self, context: 'Context', dt: 'DefDataType', enum: 'DefDataType.EnumConstructors',
                  obj: 'Any') -> 'Any':
        msg = get_value(obj, 'enum', value_pb2.Enum)
        return context.value_validate_enum(msg.constructor, enum)

    def prim_unit(self, context: 'Context', obj: 'Any') -> 'Any':
        return {}

    def prim_bool(self, context: 'Context', obj: 'Any') -> 'Any':
        return get_value(obj, 'bool', bool)

    def prim_int64(self, context: 'Context', obj: 'Any') -> 'Any':
        return get_value(obj, 'int64', int)

    def prim_text(self, context: 'Context', obj: 'Any') -> 'Any':
        return get_value(obj, 'text', str)

    def prim_timestamp(self, context: 'Context', obj: 'Any') -> 'Any':
        return to_datetime(get_value(obj, 'timestamp', timestamp_pb2.Timestamp))

    def prim_party(self, context: 'Context', obj: 'Any') -> 'Any':
        return to_party(get_value(obj, 'party', str))

    def prim_list(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'list', value_pb2.List)
        return context.convert_list(item_type, msg.elements)

    def prim_date(self, context: 'Context', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'date', str)
        return to_date(msg)

    def prim_contract_id(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'contract_id', str)
        return context.convert_contract_id(item_type, msg)

    def prim_optional(self, context: 'Context', t: 'Type', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'optional', value_pb2.Optional)
        maybe_val = msg.value if msg.HasField('value') else None
        return context.convert_optional(t, maybe_val)

    def prim_text_map(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'map', value_pb2.Map)
        mapping = {entry_pb.key: entry_pb.value for entry_pb in msg.entries}
        return context.convert_text_map(item_type, mapping)

    def prim_numeric(self, context: 'Context', nat: int, obj: 'Any') -> 'Any':
        msg = get_value(obj, 'numeric', str)
        return to_decimal(msg)

    def prim_gen_map(self, context: 'Context', key_type: 'Type', value_type: 'Type', obj: 'Any') -> 'Any':
        msg = get_value(obj, 'gen_map', value_pb2.GenMap)

        obj = {}
        for i, entry in enumerate(msg.entries):
            key = context.append_path(f'[key: {i}]').convert(key_type, entry.key)
            value = context.append_path(f'[{key}]').convert(value_type, entry.value)
            obj[key] = value

        return obj


class ProtobufEncoder(ValueMapper):
    def data_record(
            self, context: 'Context', dt: 'DefDataType', record: 'DefDataType.Fields', obj: 'Any') -> 'Any':
        msg = value_pb2.Record()

        for fld in record.fields:
            entry = msg.fields.add()
            entry.label = fld.field
            ctor, val = context.append_path(fld.field).convert(fld.type, obj[fld.field])
            set_value(entry.value, ctor, val)

        return 'record', msg

    def data_variant(
            self,
            context: 'Context',
            dt: 'DefDataType',
            variant: 'DefDataType.Fields',
            obj: 'Any') -> 'Any':
        obj_ctor, obj_value = to_variant(obj)
        obj_type = find_variant_type(dt, variant, obj_ctor)

        msg_case, msg_value = context.append_path(obj_ctor).convert(obj_type, obj_value)

        msg = value_pb2.Variant()
        msg.constructor = obj_ctor
        set_value(msg.value, msg_case, msg_value)

        return 'variant', msg

    def data_enum(self, context: 'Context', dt: 'DefDataType', enum: 'DefDataType.EnumConstructors',
                  obj: 'Any') -> 'Any':
        msg = value_pb2.Enum()
        msg.constructor = context.value_validate_enum(obj, enum)
        return 'enum', msg

    def prim_unit(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'unit', Empty()

    def prim_bool(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'bool', to_bool(obj)

    def prim_int64(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'int64', to_int(obj)

    def prim_text(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'text', to_str(obj)

    def prim_timestamp(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'timestamp', datetime_to_epoch_microseconds(to_datetime(obj))

    def prim_party(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'party', to_str(obj)

    def prim_list(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = value_pb2.List()
        for i, item in enumerate(obj):
            value = msg.elements.add()
            ctor, val = context.append_path(f'[{i}').convert(item_type, item)
            set_value(value, ctor, val)
        return 'list', msg

    def prim_date(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'date', date_to_int(to_date(obj))

    def prim_contract_id(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        return 'contract_id', to_str(obj)

    def prim_optional(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = value_pb2.Optional()
        if obj is not None:
            ctor, val = context.append_path('?').convert(item_type, obj)
            set_value(msg.value, ctor, val)
        return 'optional', msg

    def prim_text_map(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        msg = value_pb2.Map()
        for key, value in obj.items():
            entry = msg.entries.add()
            entry.key = key
            ctor, val = context.append_path(f'[{key}]').convert(item_type, value)
            set_value(entry.value, ctor, val)
        return 'map', msg

    def prim_numeric(self, context: 'Context', nat: int, obj: 'Any') -> 'Any':
        return 'numeric', decimal_to_str(to_decimal(obj))

    def prim_gen_map(
            self, context: 'Context', key_type: 'Type', value_type: 'Type', obj: 'Any') -> 'Any':
        msg = value_pb2.GenMap()
        for i, (key, value) in enumerate(obj.items()):
            entry = msg.entries.add()
            key_ctor, key_val = context.append_path(f'[key: {i}]').convert(key_type, key)
            val_ctor, val_val = context.append_path(f'[{key}]').convert(value_type, value)
            set_value(entry.key, key_ctor, key_val)
            set_value(entry.value, val_ctor, val_val)
        return 'gen_map', msg


def get_value(obj: 'Any', field: str, pb_type: 'Type[T]') -> 'T':
    return obj if isinstance(obj, pb_type) else getattr(obj, field)


def set_value(message: 'value_pb2.Value', ctor: 'Optional[str]', value: 'Any') -> None:
    """
    Work around the somewhat crazy API of Python's gRPC library to apply a known value to a
    :class:`Value`.

    :param message:
        The :class:`Value` object to modify.
    :param ctor:
        The actual field to apply to, or ``None`` to interpret the entire message as a ``Record``
        instead.
    :param value:
        The actual value to set. Must be compatible with the appropriate field.
    """
    try:
        if ctor is None:
            message.MergeFrom(value)
        elif ctor == 'unit':
            message.unit.SetInParent()
        elif ctor in ('record', 'variant', 'list', 'optional', 'enum', 'map', 'gen_map'):
            getattr(message, ctor).MergeFrom(value)
        else:
            setattr(message, ctor, value)
    except:  # noqa
        from .. import LOG
        LOG.error('Failed to set a value %s, %s', ctor, value)
        raise



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Types that describe the behavior of the ledger itself.
"""
import warnings

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .writing import Serializer


if TYPE_CHECKING:
    from ..client.pkg_loader import PackageLoader
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', DeprecationWarning)
        from .types_store import PackageStore


@dataclass(init=False, frozen=True)
class LedgerMetadata:
    """
    Attributes that are invariant with respect to any party on the ledger.
    """
    ledger_id: str
    package_loader: 'PackageLoader'
    serializer: 'Serializer'
    protocol_version: str
    _store: 'PackageStore'

    def __init__(self, ledger_id, package_loader, serializer, protocol_version, store):
        object.__setattr__(self, 'ledger_id', ledger_id)
        object.__setattr__(self, 'package_loader', package_loader)
        object.__setattr__(self, 'serializer', serializer)
        object.__setattr__(self, 'protocol_version', protocol_version)
        object.__setattr__(self, '_store', store)

    @property
    def store(self) -> 'PackageStore':
        if self._store is None:
            raise Exception('eager_package_fetch is disabled, which disables the PackageStore')

        warnings.warn('PackageStore is deprecated; use SymbolLookup '
                      '(accessible from Network.lookup) instead.',
                      DeprecationWarning, stacklevel=2)
        return self._store



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Conversion methods to Ledger API Protobuf-generated types from dazl/Pythonic types.
"""
import warnings
from typing import Any, Union

# noinspection PyPep8Naming
from . import model as G
from ...damlast.daml_lf_1 import TypeConName
from ...model.writing import AbstractSerializer, CommandPayload
from ...prim import ContractId, timedelta_to_duration
from ...values.protobuf import ProtobufEncoder, set_value

with warnings.catch_warnings():
    warnings.simplefilter('ignore', DeprecationWarning)
    from ...model.types import TypeReference


def as_identifier(tref: 'Union[TypeReference, TypeConName]') -> 'G.Identifier':
    if isinstance(tref, TypeReference):
        warnings.warn(
            'as_identifier(TypeReference) is deprecated; use as_identifier(TypeConName) instead.',
            DeprecationWarning, stacklevel=2)

        tref = tref.con

    if isinstance(tref, TypeConName):
        identifier = G.Identifier()
        _set_template(identifier, tref)
        return identifier

    else:
        raise TypeError('as_identifier requires a TypeConName or a TypeReference')


class ProtobufSerializer(AbstractSerializer):

    mapper = ProtobufEncoder()

    ################################################################################################
    # COMMAND serializers
    ################################################################################################

    def serialize_command_request(self, command_payload: CommandPayload) -> G.SubmitRequest:
        commands = [self.serialize_command(command) for command in command_payload.commands]
        return G.SubmitRequest(commands=G.Commands(
            ledger_id=command_payload.ledger_id,
            workflow_id=command_payload.workflow_id,
            application_id=command_payload.application_id,
            command_id=command_payload.command_id,
            party=command_payload.party,
            commands=commands,
            deduplication_time=(timedelta_to_duration(command_payload.deduplication_time)
                                if command_payload.deduplication_time is not None else None)))

    def serialize_create_command(self, name: 'TypeConName', template_args: 'Any') -> G.Command:
        create_ctor, create_value = template_args
        if create_ctor != 'record':
            raise ValueError('Template values must resemble records')

        cmd = G.CreateCommand()
        _set_template(cmd.template_id, name)
        set_value(cmd.create_arguments, None, create_value)
        return G.Command(create=cmd)

    def serialize_exercise_command(
            self, contract_id: 'ContractId', choice_name: str, choice_args: 'Any') \
            -> G.Command:
        type_ref = contract_id.value_type
        ctor, value = choice_args

        cmd = G.ExerciseCommand()
        _set_template(cmd.template_id, type_ref)
        cmd.contract_id = contract_id.value
        cmd.choice = choice_name
        set_value(cmd.choice_argument, ctor, value)
        return G.Command(exercise=cmd)

    def serialize_exercise_by_key_command(
            self, template_name: 'TypeConName', key_arguments: Any,
            choice_name: str, choice_arguments: Any) -> G.Command:
        key_ctor, key_value = key_arguments
        choice_ctor, choice_value = choice_arguments

        cmd = G.ExerciseByKeyCommand()
        _set_template(cmd.template_id, template_name)
        set_value(cmd.contract_key, key_ctor, key_value)
        cmd.choice = choice_name
        set_value(cmd.choice_argument, choice_ctor, choice_value)
        return G.Command(exerciseByKey=cmd)

    def serialize_create_and_exercise_command(
            self, template_name: 'TypeConName', create_arguments: 'Any',
            choice_name: str, choice_arguments: Any) -> G.Command:
        create_ctor, create_value = create_arguments
        if create_ctor != 'record':
            raise ValueError('Template values must resemble records')
        choice_ctor, choice_value = choice_arguments

        cmd = G.CreateAndExerciseCommand()
        _set_template(cmd.template_id, template_name)
        set_value(cmd.create_arguments, None, create_value)
        cmd.choice = choice_name
        set_value(cmd.choice_argument, choice_ctor, choice_value)
        return G.Command(createAndExercise=cmd)


def _set_template(message: G.Identifier, name: 'TypeConName') -> None:
    from ...damlast.util import package_ref, module_name, module_local_name
    message.package_id = package_ref(name)
    message.module_name = str(module_name(name))
    message.entity_name = module_local_name(name)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
DAML-LF fast lookups
--------------------
"""

# Implementation notes:
#
# The code in here is fairly monotonous and boilerplate heavy. However, being too creative here can
# potentially lead to performance degradations, particularly at application startup where type and
# template lookups by name are very frequent. Please be conscious of the runtime costs of
# modifications in this file!

import threading
from types import MappingProxyType
from typing import Any, Collection, Dict, NoReturn, Iterable, Optional, Tuple, AbstractSet

from .daml_lf_1 import Archive, DefDataType, DefValue, DefTemplate, DottedName, ModuleRef, \
    PackageRef, TemplateChoice, TypeConName, ValName, Package
from .errors import NameNotFoundError, PackageNotFoundError
from .protocols import SymbolLookup
from ..model.lookup import validate_template

__all__ = [
    'find_choice', 'parse_type_con_name', 'EmptyLookup', 'PackageLookup', 'MultiPackageLookup'
]


def parse_type_con_name(val: str) -> 'TypeConName':
    """
    Parse the given string as a type constructor.
    """
    # TODO: validate_template should be deprecated and some of its remaining bits be moved in-line
    #  but it's used too heavily in its current form at the moment. This is a local import to avoid
    #  import cycles between dazl.damlast and dazl.model.
    from ..model.lookup import validate_template

    pkg, name = validate_template(val)
    module_name, _, entity_name = name.rpartition(':')
    module_ref = ModuleRef(pkg, DottedName(module_name.split('.')))
    return TypeConName(module_ref, entity_name.split('.'))


def empty_lookup_impl(ref: 'Any') -> 'NoReturn':
    pkg, _ = validate_template(ref)
    if pkg != '*':
        raise PackageNotFoundError(pkg)
    else:
        raise NameNotFoundError(ref)


class EmptyLookup(SymbolLookup):
    """
    A :class:`SymbolLookup` that trivially throws for all of its functions.

    This can be used where a :class:`SymbolLookup` instance is useful but an implementation is not
    required.

    All methods are implemented such that if the provided ref has a package ID,
    :class:`PackageNotFoundError` is thrown; otherwise, :class:`NameNotFoundError` is thrown.
    """

    __slots__ = ()

    def data_type_name(self, ref: 'Any') -> 'NoReturn':
        return empty_lookup_impl(ref)

    def data_type(self, ref: 'Any') -> 'NoReturn':
        return empty_lookup_impl(ref)

    def value(self, ref: 'Any') -> 'NoReturn':
        return empty_lookup_impl(ref)

    def template_name(self, ref: 'Any') -> 'NoReturn':
        return empty_lookup_impl(ref)

    def template(self, ref: 'Any') -> 'NoReturn':
        return empty_lookup_impl(ref)


class PackageLookup(SymbolLookup):
    """
    Caching structure to make lookups on type names within a :class:`Package` faster.
    """

    def __init__(self, archive: 'Archive'):
        self.archive = archive

        data_types = {}  # type: Dict[str, Tuple[TypeConName, DefDataType]]
        values = {}  # type: Dict[str, Tuple[ValName, DefValue]]
        templates = {}  # type: Dict[str, Tuple[TypeConName, DefTemplate]]
        for module in self.archive.package.modules:
            module_ref = ModuleRef(archive.hash, module.name)

            for dt in module.data_types:
                dt_name = TypeConName(module_ref, dt.name.segments)
                data_types[f'{module.name}:{dt.name}'] = (dt_name, dt)

            for value in module.values:
                value_name = ValName(module_ref, value.name_with_type.name)
                values[f'{module.name}:{value.name_with_type.name}'] = (value_name, value)

            for tmpl in module.templates:
                tmpl_name = TypeConName(module_ref, tmpl.tycon.segments)
                templates[f'{module.name}:{tmpl.tycon}'] = (tmpl_name, tmpl)

        self._data_types = MappingProxyType(data_types)
        self._values = MappingProxyType(values)
        self._templates = MappingProxyType(templates)

    def archives(self) -> 'Collection[Archive]':
        return [self.archive]

    def package_ids(self) -> 'AbstractSet[PackageRef]':
        return frozenset([self.archive.hash])

    def data_type_name(self, ref: 'Any') -> 'TypeConName':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            dt_name = self.local_data_type_name(name)
            if dt_name is not None:
                return dt_name

        raise NameNotFoundError(ref)

    def data_type(self, ref: 'Any') -> 'DefDataType':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            dt = self.local_data_type(name)
            if dt is not None:
                return dt

        raise NameNotFoundError(ref)

    def local_data_type_name(self, name: str) -> 'Optional[TypeConName]':
        r = self._data_types.get(name)
        return r[0] if r is not None else None

    def local_data_type(self, name: str) -> 'Optional[DefDataType]':
        """
        Variation of :meth:`data_type` that assumes the name is already scoped to this package.
        Unlike :meth:`data_type`, this method returns ``None`` in the case of no match.

        You should not normally use this method directly, and instead prefer to use the methods of
        the :class:`SymbolLookup` protocol.

        :param name:
            A name to search for. Must be of the form ``"ModuleName:EntityName"``, where both
            modules and entities are dot-delimited.
        :return:
            Either a matching :class:`DefDataType`, or ``None`` if no match.
        """
        r = self._data_types.get(name)
        return r[1] if r is not None else None

    def value(self, ref: 'Any') -> 'DefValue':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            dt = self.local_value(name)
            if dt is not None:
                return dt

        raise NameNotFoundError(ref)

    def local_value(self, name: str) -> 'Optional[DefValue]':
        """
        Variation of :meth:`data_type` that assumes the name is already scoped to this package.
        Unlike :meth:`data_type`, this method returns ``None`` in the case of no match.

        You should not normally use this method directly, and instead prefer to use the methods of
        the :class:`SymbolLookup` protocol.

        :param name:
            A name to search for. Must be of the form ``"ModuleName:EntityName"``, where both
            modules and entities are dot-delimited.
        :return:
            Either a matching :class:`DefDataType`, or ``None`` if no match.
        """
        r = self._values.get(name)
        return r[1] if r is not None else None

    def template_names(self, ref: 'Any') -> 'Collection[TypeConName]':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            if name == '*':
                return self.local_template_names()
            elif name in self._templates:
                n, _ = self._templates.get(name)
                return n
        return []

    def template_name(self, ref: 'Any') -> 'TypeConName':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            tmpl = self.local_template_name(name)
            if tmpl is not None:
                return tmpl

        raise NameNotFoundError(ref)

    def template(self, ref: 'Any') -> 'DefTemplate':
        pkg, name = validate_template(ref)
        if pkg == self.archive.hash or pkg == '*':
            tmpl = self.local_template(name)
            if tmpl is not None:
                return tmpl

        raise NameNotFoundError(ref)

    def local_template_names(self) -> 'Collection[TypeConName]':
        return [n for n, _ in self._templates.values()]

    def local_template_name(self, name: str) -> 'Optional[TypeConName]':
        r = self._templates.get(name)
        return r[0] if r is not None else None

    def local_template(self, name: str) -> 'Optional[DefTemplate]':
        """
        Variation of :meth:`data_type` that assumes the name is already scoped to this package.
        Unlike :meth:`data_type`, this method returns ``None`` in the case of no match.

        You should not normally use this method directly, and instead prefer to use the methods of
        the :class:`SymbolLookup` protocol.

        :param name:
            A name to search for. Must be of the form ``"ModuleName:EntityName"``, where both
            modules and entities are dot-delimited.
        :return:
            Either a matching :class:`DefDataType`, or ``None`` if no match.
        """
        r = self._templates.get(name)
        return r[1] if r is not None else None


class MultiPackageLookup(SymbolLookup):
    """
    Combines lookups across multiple archives.
    """

    def __init__(self, archives: 'Optional[Collection[Archive]]' = None):
        self._lock = threading.Lock()
        self._cache = {}  # type: Dict[PackageRef, PackageLookup]
        if archives is not None:
            self.add_archive(*archives)

    def archives(self) -> 'Collection[Archive]':
        """
        Return the list of known archives.
        """
        return [lookup.archive for lookup in self._cache.values()]

    def add_archive(self, *a: 'Archive') -> None:
        """
        Add one or more :class:`Archive`s to this lookup.

        This method is thread-safe, but note that :class:`MultiPackageLookup` allows dirty reads for
        performance reasons.

        :param a: One or more :class:`Archive`s to add.
        """
        new_lookups = {ar.hash: PackageLookup(ar) for ar in a}
        with self._lock:
            # replace the old cache with a new one, incorporating values from both the existing
            # cache and the new lookups that came in. When there is a key conflict between the two
            # APIs, always prefer the existing cache to provide stability to callers.
            self._cache = {**new_lookups, **self._cache}

    def package_ids(self) -> 'AbstractSet[PackageRef]':
        return set(self._cache)

    def package(self, ref: 'PackageRef') -> 'Package':
        lookup = self._cache.get(ref)
        if lookup is not None:
            return lookup.archive.package

        raise PackageNotFoundError(ref)

    def data_type_name(self, ref: 'Any') -> 'TypeConName':
        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            dt_name = lookup.local_data_type_name(name)
            if dt_name is not None:
                return dt_name

        raise NameNotFoundError(ref)

    def data_type(self, ref: 'Any') -> 'DefDataType':
        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            dt = lookup.local_data_type(name)
            if dt is not None:
                return dt

        raise NameNotFoundError(ref)

    def value(self, ref: 'Any') -> 'DefValue':
        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            val = lookup.local_value(name)
            if val is not None:
                return val

        raise NameNotFoundError(ref)

    def template_names(self, ref: 'Any') -> 'Collection[TypeConName]':
        names = []

        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            if name == '*':
                names.extend(lookup.local_template_names())
            else:
                n = lookup.local_template_name(name)
                if n is not None:
                    names.append(n)

        return names

    def template_name(self, ref: 'Any') -> 'TypeConName':
        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            n = lookup.local_template_name(name)
            if n is not None:
                return n

        raise NameNotFoundError(ref)

    def template(self, ref: 'Any') -> 'DefTemplate':
        pkg, name = validate_template(ref)
        for lookup in self._lookups(pkg):
            tmpl = lookup.local_template(name)
            if tmpl is not None:
                return tmpl

        raise NameNotFoundError(ref)

    def _lookups(self, ref: 'PackageRef') -> 'Iterable[PackageLookup]':
        """
        Return the individual :class:`PackageLookup` objects that should be consulted based on the
        :class:`PackageRef`.

        :param ref:
            A :class:`PackageRef` to look for.
        :return:
            A collection of :class:`PackageLookup` objects that match the :class:`PackageRef`. This
            collection is never empty.
        :raises PackageNotFoundError:
            if the :class:`PackageRef` points to a package that is not present in this lookup.
        """
        if ref == '*':
            return self._cache.values()

        lookup = self._cache.get(ref)
        if lookup is not None:
            return lookup,

        raise PackageNotFoundError(ref)


def find_choice(template: 'DefTemplate', name: str) -> 'TemplateChoice':
    """
    Find a choice in a :class:`DefTemplate`. If the choice could not be found,
    :class:`NameNotFoundError` is raised.
    """
    for choice in template.choices:
        if choice.name == name:
            return choice

    raise NameNotFoundError(f'choice {name}')



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import warnings
from asyncio import Future, sleep
from collections import defaultdict
from typing import Awaitable, Dict, List, Optional, Union, Collection, cast

from ..client._reader_match import is_match
from ..damlast.daml_lf_1 import TypeConName
from ..damlast.protocols import SymbolLookup
from ..model.core import ContractId, ContractsState, ContractMatch, \
    UnknownTemplateWarning, ContractContextualData, ContractContextualDataCollection
from ..model.reading import ContractCreateEvent, ContractArchiveEvent
from ..scheduler import Invoker
from ..util.asyncio_util import await_then


class ActiveContractSet:
    def __init__(self, invoker: 'Invoker', lookup: 'SymbolLookup'):
        self.invoker = invoker
        self.lookup = lookup
        self._tcdata = defaultdict(TemplateContractData)  # type: Dict[TypeConName, TemplateContractData]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._tcdata[event.cid.value_type].handle_create(event)

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        self._tcdata[event.cid.value_type].handle_archive(event)

    def get(self, cid: 'Union[str, ContractId]') -> 'Optional[ContractContextualData]':
        """
        Return information for the associated :class:`ContractId`.

        :param cid: Either a ContractId or a string that represents a ContractId.
        :return: Contract information for the specified contract ID.
        """
        tcd = self._tcdata.get(cid.value_type) if isinstance(cid, ContractId) else None
        if tcd is not None:
            return tcd.get(cid)
        else:
            # untyped ContractId lookup, so scan across all templates
            for tcd in self._tcdata.values():
                cxd = tcd.get(cid)
                if cxd is not None:
                    return cxd
            return None

    def read_active(self, template_name: str, match: ContractMatch = None) -> ContractsState:
        return {cxd.cid: cxd.cdata for cxd in self.read_full(template_name, match)}

    def read_full(self,
                  template_name: str,
                  match: ContractMatch = None,
                  include_archived: bool = False) -> ContractContextualDataCollection:
        unfiltered = self._get_template_state(template_name)
        all_items = [item
                     for tcd in unfiltered.values()
                     for item in tcd.subset(match, include_archived)]
        return ContractContextualDataCollection(all_items)

    async def read_async(
            self, template_name: str, match: ContractMatch = None, min_count: int = 1) \
            -> 'ContractsState':
        from .. import LOG

        # Fetch matching names; we may need to wait for a package to be fetched in order to do this.
        # Note that this loop may never terminate; in this case, this behavior is actually
        # acceptable because it is what enables us to wait indefinitely for the creation of a
        # contract of a specific type. Because we need type information available to parse data
        # from the read stream, by the time data ends up in the ACS, the lookup will definitely
        # know of the template name.
        names = []
        fail_count = 0
        while not names:
            names = self.lookup.template_names(template_name)
            if not names:
                fail_count += 1
                await sleep(1)

                if fail_count % 15 == 0:
                    LOG.warning(
                        'ACS read_async(%r) is still waiting for a package that contains that '
                        'template. If this is taking longer than you expect, it might be because '
                        'the relevant package is not loaded on your ledger.', template_name)

        unfiltered = self._get_template_state(template_name)
        if len(unfiltered) > 1:
            warnings.warn('Wildcard searches are not supported on async ACS queries',
                          UnknownTemplateWarning, stacklevel=3)
            unfiltered = dict()  # type: Dict[TypeConName, TemplateContractData]

        (tt, tcd), = unfiltered.items()

        query = PendingQuery(self.invoker, match, min_count)
        # if the current state is already a match, then don't remember the query since we're
        # already done
        if tcd is None or not query.check_ready(tcd):
            tcd.register_query(query)

        return await await_then(query.future, lambda cxds: {cxd.cid: cxd.cdata for cxd in cxds})

    def _get_template_state(self, template_name: str) -> 'Dict[TypeConName, TemplateContractData]':
        names = self.lookup.template_names(template_name)
        if not names:
            warnings.warn(f'Unknown template name: {template_name}', UnknownTemplateWarning,
                          stacklevel=4)

        return {name: self._tcdata[name] for name in names}


class TemplateContractData:
    """
    Storage for state related to a specific :class:`Template`.
    """
    def __init__(self):
        # raw ContractId to ContractContextualData
        self._data = dict()  # type: Dict[str, ContractContextualData]
        # outstanding ACS queries
        self._queries = list()  # type: List[PendingQuery]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._data[event.cid.value] = ContractContextualData(
            cid=event.cid,
            cdata=event.cdata,
            effective_at=event.time,
            archived_at=None,
            active=True)

        self._queries = [q for q in self._queries if not q.check_ready(self)]

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        existing_data = self._data.get(event.cid.value)
        effective_at = existing_data.effective_at if existing_data is not None else None
        self._data[event.cid.value] = ContractContextualData(
            cid=event.cid,
            cdata=None,
            effective_at=effective_at,
            archived_at=event.time,
            active=False)

    def get(self, cid: 'Union[str, ContractId]') -> 'Optional[ContractContextualData]':
        if isinstance(cid, ContractId):
            cid = cid.value
        return self._data.get(cid)

    def subset(self, match: 'ContractMatch', include_archived: bool) -> 'Collection[ContractContextualData]':
        return [cxd for cxd in self._data.values()
                if (include_archived or cxd.active)
                and is_match(match, cxd.cdata)]

    def register_query(self, query: 'PendingQuery') -> None:
        self._queries.append(query)


class PendingQuery:
    def __init__(self, invoker: 'Invoker', match, min_count: int):
        self.future = invoker.create_future()  # type: Awaitable[Collection[ContractContextualData]]
        self.match = match
        self.min_count = min_count

    def check_ready(self, tcd: 'TemplateContractData') -> bool:
        matches = tcd.subset(self.match, False)
        if len(matches) >= self.min_count:
            cast(Future, self.future).set_result(matches)
            return True
        else:
            return False



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import warnings

warnings.warn('dazl.protocols.v0 is deprecated', DeprecationWarning, stacklevel=2)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from ..damlast.daml_lf_1 import TypeConName
from ..prim import ContractId
from ..model.writing import CreateCommand, ExerciseCommand, ExerciseByKeyCommand, \
    CreateAndExerciseCommand, AbstractSerializer
from ..values import CanonicalMapper


class ValidateSerializer(AbstractSerializer):
    """
    A serializer that doesn't actually write to an on-wire form, but merely does lightweight
    conversion and give on-wire serialization the best chance of succeeding.

    This class essentially handles type conversion before a command is fully created and sent to
    a background thread for eventual writing to the Ledger API.
    """

    mapper = CanonicalMapper()

    def serialize_create_command(self, name: 'TypeConName', template_args: 'Any') \
            -> 'CreateCommand':
        return CreateCommand(template=name, arguments=template_args)

    def serialize_exercise_command(
            self, contract_id: 'ContractId', choice_name: str, choice_args: 'Any') \
            -> 'ExerciseCommand':
        return ExerciseCommand(contract=contract_id, choice=choice_name, arguments=choice_args)

    def serialize_exercise_by_key_command(
            self, template_name: 'TypeConName', key_arguments: Any,
            choice_name: str, choice_arguments: 'Any') -> 'ExerciseByKeyCommand':
        return ExerciseByKeyCommand(
            template=template_name, contract_key=key_arguments,
            choice=choice_name, choice_argument=choice_arguments)

    def serialize_create_and_exercise_command(
            self, template_name: 'TypeConName', create_arguments: Any,
            choice_name: str, choice_arguments: 'Any') -> 'CreateAndExerciseCommand':
        return CreateAndExerciseCommand(
            template=template_name, arguments=create_arguments,
            choice=choice_name, choice_argument=choice_arguments)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Conversion methods from Ledger API Protobuf-generated types to dazl/Pythonic types.
"""
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Sequence, Union

from ... import LOG
from ..._gen.com.daml.ledger.api.v1 import \
    active_contracts_service_pb2 as acs_pb2, \
    event_pb2, \
    ledger_offset_pb2 as lo_pb2, \
    transaction_pb2 as tx_pb2, \
    transaction_filter_pb2 as txf_pb2, \
    transaction_service_pb2 as txs_pb2, \
    value_pb2
from ...damlast.daml_lf_1 import ModuleRef, PackageRef, DottedName, TypeConName
from ...damlast.daml_types import con
from ...damlast.lookup import find_choice
from ...damlast.protocols import SymbolLookup
from ...model.reading import BaseEvent, TransactionFilter, ContractCreateEvent, \
    TransactionStartEvent, TransactionEndEvent, ContractArchiveEvent, OffsetEvent, \
    ContractExercisedEvent, ActiveContractSetEvent, ContractFilter
from ...prim import Party, to_datetime
from ...values import Context, ProtobufDecoder

with warnings.catch_warnings():
    warnings.simplefilter('ignore', DeprecationWarning)
    from ...model.types_store import PackageStore


DECODER = ProtobufDecoder()


@dataclass(frozen=True)
class BaseEventDeserializationContext:
    """
    Attributes required throughout the deserialization of an event stream.
    """
    client: 'Any'
    lookup: 'SymbolLookup'
    store: 'PackageStore'
    party: 'Party'
    ledger_id: str

    def deserializer_context(self) -> 'Context':
        return Context(DECODER, self.lookup)

    def offset_event(self, time: datetime, offset: str) -> OffsetEvent:
        return OffsetEvent(
            self.client, self.party, time, self.ledger_id, self.lookup, self.store, offset)

    def active_contract_set(self, offset: str, workflow_id: str) \
            -> 'ActiveContractSetEventDeserializationContext':
        return ActiveContractSetEventDeserializationContext(
            self.client, self.lookup, self.store, self.party, self.ledger_id,
            offset, workflow_id)

    def transaction(self, time: datetime, offset: str, command_id: str, workflow_id: str) \
            -> 'TransactionEventDeserializationContext':
        return TransactionEventDeserializationContext(
            self.client, self.lookup, self.store, self.party, self.ledger_id,
            time, offset, command_id, workflow_id)


@dataclass(frozen=True)
class ActiveContractSetEventDeserializationContext(BaseEventDeserializationContext):
    """
    Attributes required throughout the deserialization of the Active Contract Set.
    """
    offset: str
    workflow_id: str

    def active_contract_set_event(self, contract_events: 'Sequence[ContractCreateEvent]') \
            -> 'ActiveContractSetEvent':
        return ActiveContractSetEvent(
            self.client, self.party, None, self.ledger_id, self.lookup, self.store, self.offset,
            contract_events)

    def contract_created_event(self, cid, cdata, event_id, witness_parties) -> ContractCreateEvent:
        return ContractCreateEvent(
            client=self.client, party=self.party, time=None, ledger_id=self.ledger_id,
            lookup=self.lookup, package_store=self.store, offset=self.offset,
            command_id='', workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)


@dataclass(frozen=True)
class TransactionEventDeserializationContext(BaseEventDeserializationContext):
    """
    Attributes required throughout the deserialization of an event stream.
    """
    time: datetime
    offset: str
    command_id: str
    workflow_id: str

    def transaction_start_event(self, contract_events) -> TransactionStartEvent:
        return TransactionStartEvent(
            self.client, self.party, self.time, self.ledger_id, self.lookup, self.store,
            self.offset, self.command_id, self.workflow_id, contract_events)

    def transaction_end_event(self, contract_events) -> TransactionEndEvent:
        return TransactionEndEvent(
            self.client, self.party, self.time, self.ledger_id, self.lookup, self.store,
            self.offset, self.command_id, self.workflow_id, contract_events)

    def contract_created_event(self, cid, cdata, event_id, witness_parties) -> ContractCreateEvent:
        return ContractCreateEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, lookup=self.lookup, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)

    def contract_exercised_event(self, cid, cdata, event_id, witness_parties,
                                 contract_creating_event_id: str, choice: str, choice_argument: Any,
                                 acting_parties, consuming, child_event_ids, exercise_result) \
            -> ContractExercisedEvent:
        return ContractExercisedEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, lookup=self.lookup, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties,
            contract_creating_event_id=contract_creating_event_id, acting_parties=acting_parties,
            choice=choice, choice_args=choice_argument, consuming=consuming,
            child_event_ids=child_event_ids,
            exercise_result=exercise_result)

    def contract_archived_event(self, cid, cdata, event_id, witness_parties) \
            -> ContractArchiveEvent:
        return ContractArchiveEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, lookup=self.lookup, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)


def serialize_transactions_request(f: 'TransactionFilter', ledger_id: str, party: str) \
        -> 'txs_pb2.GetTransactionsRequest':
    if f.current_offset is not None:
        ledger_offset = lo_pb2.LedgerOffset()
        ledger_offset.absolute = f.current_offset
    else:
        ledger_offset = lo_pb2.LedgerOffset()
        ledger_offset.boundary = 0

    if f.destination_offset is not None:
        final_offset = lo_pb2.LedgerOffset()
        final_offset.absolute = f.destination_offset
    else:
        final_offset = lo_pb2.LedgerOffset()
        final_offset.boundary = 1

    return txs_pb2.GetTransactionsRequest(
        ledger_id=ledger_id,
        begin=ledger_offset,
        end=final_offset,
        filter=serialize_transaction_filter(f, party))


def serialize_acs_request(f: 'ContractFilter', ledger_id: str, party: str) \
        -> 'acs_pb2.GetActiveContractsRequest':
    return acs_pb2.GetActiveContractsRequest(
        ledger_id=ledger_id,
        filter=serialize_transaction_filter(f, party))


def serialize_event_id_request(ledger_id: str, event_id: str, requesting_parties: 'Sequence[str]') \
        -> 'txs_pb2.GetTransactionByEventIdRequest':
    return txs_pb2.GetTransactionByEventIdRequest(
        ledger_id=ledger_id,
        event_id=event_id,
        requesting_parties=requesting_parties)


def serialize_transaction_filter(contract_filter: 'ContractFilter', party: str) \
        -> 'txf_pb2.TransactionFilter':
    from .pb_ser_command import as_identifier

    identifiers = [as_identifier(tt) for tt in contract_filter.templates] \
        if contract_filter.templates is not None else None

    parties = [party]
    if contract_filter.party_groups is not None:
        parties.extend(contract_filter.party_groups)

    filters_by_party = {}
    for party in parties:
        if identifiers is not None:
            filters_by_party[party] = txf_pb2.Filters(
                inclusive=txf_pb2.InclusiveFilters(template_ids=identifiers))
        else:
            filters_by_party[party] = txf_pb2.Filters()

    return txf_pb2.TransactionFilter(filters_by_party=filters_by_party)


def to_transaction_events(
        context: 'BaseEventDeserializationContext',
        tx_stream_pb: 'Iterable[txs_pb2.GetTransactionsResponse]',
        tt_stream_pb: 'Optional[Iterable[txs_pb2.GetTransactionTreesResponse]]',
        last_offset_override: 'Optional[str]') -> 'Sequence[BaseEvent]':
    """
    Convert a stream of :class:`GetTransactionsResponse` into a sequence of events.

    :param context:
        Additional data required for deserializing wire information into easily-consumable
        application events.
    :param tx_stream_pb:
        A stream of :class:`GetTransactionsResponse`.
    :param tt_stream_pb:
        An optional stream of :class:`GetTransactionTreesResponse`, used to enrich the stream with
        exercise events. Transactions included in this message that are _not_ in ``tx_stream_pb``
        will be discarded.
    :param last_offset_override:
        An optional last offset, that, if specified AND different from the last offset of the
        stream, causes a synthetic :class:`OffsetEvent` to be created at the end.
    :return:
    """
    events = []  # type: List[OffsetEvent]

    events_by_offset = dict()  # type: Dict[str, List[OffsetEvent]]
    for item_pb in tx_stream_pb:
        for transaction_pb in item_pb.transactions:
            events_by_offset[transaction_pb.offset] = list(
                to_transaction_chunk(context, transaction_pb))

    if tt_stream_pb is not None:
        for transaction_tree_pb in tt_stream_pb:
            for transaction_pb in transaction_tree_pb.transactions:
                tx_events = events_by_offset.get(transaction_pb.offset)
                if tx_events is not None:
                    tx_events[-1:-1] = from_transaction_tree(context, transaction_pb)

    for tx_events in events_by_offset.values():
        events.extend(tx_events)

    if last_offset_override is not None:
        last_time = None if not events else events[-1].time
        last_offset = None if not events else events[-1].offset
        if last_offset != last_offset_override:
            events.append(context.offset_event(last_time, last_offset_override))

    return events


def to_acs_events(
        context: 'BaseEventDeserializationContext',
        acs_stream_pb: 'Iterable[Any]') -> 'Sequence[ActiveContractSetEvent]':
    return [acs_evt
            for acs_response_pb in acs_stream_pb
            for acs_evt in to_acs_event(context, acs_response_pb)]


def from_transaction_tree(
        context: 'BaseEventDeserializationContext', tt_pb: 'tx_pb2.TransactionTree') \
        -> 'Sequence[OffsetEvent]':
    t_context = context.transaction(
        time=to_datetime(tt_pb.effective_at),
        offset=tt_pb.offset,
        command_id=tt_pb.command_id,
        workflow_id=tt_pb.workflow_id)

    events = []
    for evt_pb in tt_pb.events_by_id.values():
        evt = to_event(t_context, evt_pb)
        if isinstance(evt, ContractExercisedEvent):
            events.append(evt)

    return events


def to_acs_event(
        context: 'BaseEventDeserializationContext', acs_pb: 'acs_pb2.GetActiveContractsResponse'):
    acs_context = context.active_contract_set(acs_pb.offset, acs_pb.workflow_id)
    contract_events = \
        [evt
         for evt in (to_created_event(acs_context, evt_pb) for evt_pb in acs_pb.active_contracts)
         if evt is not None]
    return [acs_context.active_contract_set_event(contract_events)]


def to_transaction_chunk(context: 'BaseEventDeserializationContext', tx_pb: 'tx_pb2.Transaction') \
        -> 'Sequence[OffsetEvent]':
    """
    Return a sequence of events parsed from a Ledger API ``Transaction`` protobuf message.

    :param context:
        Additional information taken from the enclosing message.
    :param tx_pb:
        The Ledger API ``Transaction`` to parse.
    :return:
        A :class:`TransactionStartEvent`, followed by zero or more :class:`ContractCreateEvent`
        and/or :class:`ContractArchiveEvent`, followed by :class:`TransactionEndEvent`.
    """

    t_context = context.transaction(
        time=to_datetime(tx_pb.effective_at),
        offset=tx_pb.offset,
        command_id=tx_pb.command_id,
        workflow_id=tx_pb.workflow_id)

    contract_events = [to_event(t_context, evt_pb) for evt_pb in tx_pb.events]
    contract_events = [evt for evt in contract_events if evt is not None]

    return [t_context.transaction_start_event(contract_events),
            *contract_events,
            t_context.transaction_end_event(contract_events)]


def to_event(
        context: 'Union[TransactionEventDeserializationContext, ActiveContractSetEventDeserializationContext]',
        evt_pb: 'event_pb2.Event') \
        -> 'Optional[BaseEvent]':
    try:
        event_type = evt_pb.WhichOneof('event')
    except ValueError:
        try:
            event_type = evt_pb.WhichOneof('kind')
        except ValueError:
            LOG.error('Deserialization error into an event of %r', evt_pb)
            raise

    if 'created' == event_type:
        return to_created_event(context, evt_pb.created)
    elif 'exercised' == event_type:
        return to_exercised_event(context, evt_pb.exercised)
    elif 'archived' == event_type:
        return to_archived_event(context, evt_pb.archived)
    else:
        raise ValueError(f'unknown event type: {event_type}')


def to_created_event(
        context: 'Union[TransactionEventDeserializationContext, ActiveContractSetEventDeserializationContext]',
        cr: 'event_pb2.CreatedEvent') \
        -> 'Optional[ContractCreateEvent]':
    tt = con(to_type_con_name(cr.template_id))

    ctx = context.deserializer_context()
    cid = ctx.convert_contract_id(tt, cr.contract_id)
    cdata = ctx.convert(tt, cr.create_arguments)
    event_id = cr.event_id
    witness_parties = tuple(cr.witness_parties)

    return context.contract_created_event(cid, cdata, event_id, witness_parties)


def to_exercised_event(
        context: 'TransactionEventDeserializationContext',
        er: 'event_pb2.ExercisedEvent') \
        -> 'Optional[ContractExercisedEvent]':
    name = to_type_con_name(er.template_id)
    tt = con(name)

    template = context.lookup.template(name)
    choice = find_choice(template, er.choice)

    ctx = context.deserializer_context()

    cid = ctx.convert_contract_id(tt, er.contract_id)
    event_id = er.event_id
    witness_parties = tuple(er.witness_parties)
    try:
        contract_creating_event_id = er.contract_creating_event_id
    except AttributeError:
        contract_creating_event_id = None
    choice_args = ctx.convert(choice.arg_binder.type, er.choice_argument)
    acting_parties = tuple(er.acting_parties)
    consuming = er.consuming
    child_event_ids = er.child_event_ids
    exercise_result = ctx.convert(choice.ret_type, er.exercise_result)

    return context.contract_exercised_event(
        cid, None, event_id, witness_parties, contract_creating_event_id,
        er.choice, choice_args, acting_parties, consuming, child_event_ids, exercise_result)


def to_archived_event(
        context: 'TransactionEventDeserializationContext',
        ar: 'event_pb2.ArchivedEvent') \
        -> 'Optional[ContractArchiveEvent]':
    tt = con(to_type_con_name(ar.template_id))
    event_id = ar.event_id
    witness_parties = tuple(ar.witness_parties)

    ctx = context.deserializer_context()
    cid = ctx.convert_contract_id(tt, ar.contract_id)
    return context.contract_archived_event(cid, None, event_id, witness_parties)


def to_type_con_name(identifier: 'value_pb2.Identifier') -> 'TypeConName':
    return TypeConName(
        ModuleRef(
            PackageRef(identifier.package_id),
            DottedName(identifier.module_name.split('.'))
        ),
        DottedName(identifier.entity_name.split('.')).segments)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .model import Formatter

__all__ = ['DEFAULT_FORMATTER_NAME', 'get_formatter']

DEFAULT_FORMATTER_NAME = 'pretty'


def get_formatter(fmt) -> 'Formatter':
    """
    Return a formatter of the specified name.
    """
    if fmt is None:
        fmt = DEFAULT_FORMATTER_NAME

    if fmt == 'json':
        from .fmt_json import JsonFormatter
        return JsonFormatter()
    elif fmt == 'pretty':
        from .fmt_pretty import PrettyFormatter
        return PrettyFormatter()
    else:
        raise ValueError('unknown format: {}'.format(fmt))



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from ..damlast.daml_lf_1 import Type as DamlType, DefDataType
from .mapper import ValueMapper, Context, convert


class PythonSampleEncoder(ValueMapper):
    def __init__(self, max_depth: int = 6):
        self.max_depth = max_depth

    def _recurse(self, context: 'Context', key, item_type):
        if context.depth <= self.max_depth:
            return '...'
        else:
            return convert(self, context.append_path(key), item_type, None)

    def data_record(self, context: 'Context', dt: 'DefDataType', record: 'DefDataType.Fields', obj: 'Any') -> 'Any':
        return {fld.field: self._recurse(context, fld.field, fld.type) for fld in record.fields}
        pass

    def data_variant(self, context: 'Context', dt: 'DefDataType', variant: 'DefDataType.Fields', obj: 'Any') -> 'Any':
        return {fld.field: self._recurse(context, fld.field, fld.type) for fld in variant.fields}

    def data_enum(self, context: 'Context', dt: 'DefDataType', enum: 'DefDataType.EnumConstructors', obj: 'Any') -> 'Any':
        return ' | '.join(enum.constructors)

    def prim_unit(self, context: 'Context', obj: 'Any') -> 'Any':
        return '{}'

    def prim_bool(self, context: 'Context', obj: 'Any') -> 'Any':
        return 'true | false'

    def prim_int64(self, context: 'Context', obj: 'Any') -> 'Any':
        return "int-value"

    def prim_text(self, context: 'Context', obj: 'Any') -> 'Any':
        return "text-value"

    def prim_timestamp(self, context: 'Context', obj: 'Any') -> 'Any':
        return "timestamp-value"

    def prim_party(self, context: 'Context', obj: 'Any') -> 'Any':
        return "party-value"

    def prim_list(self, context: 'Context', item_type: 'DamlType', obj: 'Any') -> 'Any':
        return [self._recurse(context, '[]', item_type)]

    def prim_date(self, context: 'Context', obj: 'Any') -> 'Any':
        return "date-value"

    def prim_contract_id(self, context: 'Context', item_type: 'DamlType', obj: 'Any') -> 'Any':
        return "contract-id"

    def prim_optional(self, context: 'Context', item_type: 'DamlType', obj: 'Any') -> 'Any':
        return ""

    def prim_text_map(self, context: 'Context', item_type: 'DamlType', obj: 'Any') -> 'Any':
        return [self._recurse(context, '[]', item_type)]

    def prim_numeric(self, context: 'Context', nat: int, obj: 'Any') -> 'Any':
        pass

    def prim_gen_map(self, context: 'Context', key_type: 'DamlType', value_type: 'DamlType', obj: 'Any') -> 'Any':
        return {self._recurse(context, '[]', key_type): self._recurse(context, '[]', value_type)}



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentParser

from .. import LOG, Network
from ._base import CliCommand
from ..pretty.table import write_acs, DEFAULT_FORMATTER_NAME
from ..client.config import configure_parser, NetworkConfig


class ListAllCommand(CliCommand):
    name = 'ls'

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser('dazl ls')

        configure_parser(arg_parser, config_file_support=True)
        arg_parser.add_argument('--format', '--fmt', '-F', type=str, default=DEFAULT_FORMATTER_NAME)
        arg_parser.add_argument('--template-filter', '-T', type=str)
        arg_parser.add_argument('--all', '-A', action='store_true')
        return arg_parser

    def execute(self, args) -> int:
        fmt = args.format
        template_filter = [template.strip() for template in args.template_filter.split(',')] \
            if args.template_filter is not None else None
        include_archived = bool(args.all)
        LOG.debug('Executing an ls...')

        final_config = NetworkConfig.get_config(args)

        network = Network()
        network.set_config(final_config)
        network.set_config(eager_package_fetch=False)
        if include_archived:
            network.set_config(use_acs_service=False)

        for party in args.parties:
            network.aio_party(party)

        network.run_until_complete(self._main(network, fmt, include_archived))
        return 0

    async def _main(self, network: 'Network', fmt: str, include_archived: bool):
        import sys

        LOG.debug('Starting our parties...')
        for party in network.parties():
            await network.aio_party(party).ready()
            LOG.debug('Party %s is ready.', party)
        metadata = await network.aio_global().metadata()
        LOG.debug('Our parties are now ready.')

        write_acs(sys.stdout, network, fmt=fmt, include_archived=include_archived)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, NewType

from .basic import to_str

__all__ = ['Party', 'to_party']

Party = NewType('Party', str)


def to_party(o: 'Any') -> 'Party':
    return Party(to_str(o))



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import AbstractSet, Any, TYPE_CHECKING, Collection

if sys.version_info >= (3, 7):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from .daml_lf_1 import DefDataType, DefTemplate, DefValue, Package, PackageRef, TypeConName, Archive

__all__ = ['PackageProvider', 'SymbolLookup']


@runtime_checkable
class PackageProvider(Protocol):
    """
    Structural interface for a class that can synchronously provide package data.

    If a requested package could not be found, :class:`PackageNotFoundError` should be raised; the
    functions of this protocol should never return ``None``.
    """

    def package(self, ref: 'PackageRef') -> 'Package':
        """
        Return the :class:`Package` for the specified :class:`PackageRef`.

        :param ref:
            The package ID to look up.
        :return:
            The Package corresponding to this package ID.
        :raises PackageNotFoundError:
            if the package could not be found
        """
        raise NotImplementedError('PackageProvider.package must be implemented')

    def package_ids(self) -> 'AbstractSet[PackageRef]':
        """
        Return all package IDs that are known to this :class:`PackageProvider`.

        Implementations are allowed to return larger sets of packages over time, but should
        generally never omit a package ID that was previously returned.

        :return:
            A set of package IDs.
        """
        raise NotImplementedError('PackageProvider.package_ids must be implemented')


@runtime_checkable
class SymbolLookup(Protocol):
    """
    Structural interface for a class that can synchronously provide type/value information.

    If a requested _package_ could not be found, :class:`PackageNotFoundError` is raised.
    If a requested package is known to this :class:`SymbolLookup` but the name is unknown, then
    :class:`NameNotFoundError` is raised. Callers can use this to disambiguate between errors that
    are retryable (missing packages can be fetched) and non-retryable (names within a known package,
    if they do not currently exist, will NEVER exist).
    """

    def archives(self) -> 'Collection[Archive]':
        """
        Return the archives that are known to this lookup.
        """
        raise NotImplementedError('SymbolLookup.archives must be implemented')

    def package_ids(self) -> 'AbstractSet[PackageRef]':
        """
        Return the package IDs that are known to this lookup.
        """
        raise NotImplementedError('SymbolLookup.package_ids must be implemented')

    def data_type_name(self, ref: 'Any') -> 'TypeConName':
        """
        Return the :class:`TypeConName` that refers to a :class:`DefDataType` that is known to
        exist in this lookup.

        If this method succeeds, :meth:`data_type` with the returned :class:`TypeConName` as an
        argument should also always succeed.
        """
        raise NotImplementedError('SymbolLookup.data_type_name must be implemented')

    def data_type(self, ref: 'Any') -> 'DefDataType':
        """
        Return the :class:`DefDataType` for the specified name.
        """
        raise NotImplementedError('SymbolLookup.data_type must be implemented')

    def value(self, ref: 'Any') -> 'DefValue':
        """
        Return the :class:`DefValue` for the specified name.
        """
        raise NotImplementedError('SymbolLookup.value must be implemented')

    def template_names(self, ref: 'Any') -> 'Collection[TypeConName]':
        """
        Return all template names that are currently known that are a match for the query. Either
        :class:`PackageRef` or the template name can be `*`.

        Unlike the other methods of this class, this function should return an empty list in the
        case of a match failure; :class:`PackageNotFoundError` or :class:`NameNotFoundError` are
        never thrown.
        """
        raise NotImplementedError('SymbolLookup.template_names must be implemented')

    def template_name(self, ref: 'Any') -> 'TypeConName':
        """
        Return the :class:`TypeConName` that refers to a :class:`DefTemplate` that is known to
        exist in this lookup.

        If this method succeeds, :meth:`data_type` with the returned :class:`TypeConName` as an
        argument should also always succeed.
        """
        raise NotImplementedError('SymbolLookup.template_name must be implemented')

    def template(self, ref: 'Any') -> 'DefTemplate':
        """
        Return the :class:`DefTemplate` for the specified name.
        """
        raise NotImplementedError('SymbolLookup.template must be implemented')



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Any

from .context import Context
from .mapper import ValueMapper

from ..damlast.daml_lf_1 import DefDataType, Type
from ..damlast.util import find_variant_type
from ..prim import to_bool, to_str, datetime_to_str, date_to_str, to_decimal, decimal_to_str, to_int, to_party, \
    to_variant

__all__ = ['FlatStringMapper', 'ArrayStringMapper']


def maybe_parenthesize(value: str) -> str:
    return '(' + value + ')' if ' ' in value else value


class StringMapperBase(ValueMapper):
    """
    Return string representations of all types in a format that looks closest to their
    representation in DAML.
    """

    def data_record(
            self, context: 'Context', dt: 'DefDataType', record: 'DefDataType.Fields', obj: 'Any') \
            -> 'Any':
        raise NotImplementedError('StringMapperBase.data_record has no implementation')

    def data_variant(
            self,
            context: 'Context',
            dt: 'DefDataType',
            variant: 'DefDataType.Fields',
            obj: 'Any') -> 'Any':
        ctor, value = to_variant(obj)
        obj_type = find_variant_type(dt, variant, ctor)

        value_str = context.append_path(ctor, mapper=FLAT).convert(obj_type, value)
        return f'{ctor} ({maybe_parenthesize(value_str)})'

    def data_enum(
            self, context: 'Context', dt: 'DefDataType', enum: 'DefDataType.EnumConstructors',
            obj: 'Any') -> 'Any':
        return context.value_validate_enum(obj, enum)

    def prim_unit(self, context: 'Context', obj: 'Any') -> 'Any':
        return "()"

    def prim_bool(self, context: 'Context', obj: 'Any') -> 'Any':
        return "true" if to_bool(obj) else "false"

    def prim_int64(self, context: 'Context', obj: 'Any') -> 'Any':
        return str(to_int(obj))

    def prim_text(self, context: 'Context', obj: 'Any') -> 'Any':
        """
        Return text formatted as a string, with double quotes.
        """
        return json.dumps(to_str(obj))

    def prim_timestamp(self, context: 'Context', obj: 'Any') -> 'Any':
        return datetime_to_str(obj)

    def prim_party(self, context: 'Context', obj: 'Any') -> 'Any':
        """
        Return a Party literal formatted as a string, with single quotes.
        """
        return repr(to_party(obj))

    def prim_list(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        return '[' + ', '.join(context.convert_list(item_type, obj, mapper=FLAT)) + ']'

    def prim_date(self, context: 'Context', obj: 'Any') -> 'Any':
        return date_to_str(obj)

    def prim_contract_id(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        return context.convert_contract_id(item_type, obj).value

    def prim_optional(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        if obj is not None:
            value_str = context.convert_optional(item_type, obj, mapper=FLAT)
            if ' ' in value_str:
                return f'Some ({value_str})'
            else:
                return f'Some {value_str}'
        else:
            return 'None'

    def prim_text_map(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        return "{" + ", ".join(
            json.dumps(key) + ": " + value
            for key, value in context.convert_text_map(item_type, obj, mapper=FLAT).items()) + "}"

    def prim_numeric(self, context: 'Context', nat: int, obj: 'Any') -> 'Any':
        return decimal_to_str(to_decimal(obj))

    def prim_gen_map(self, context: 'Context', key_type: 'Type', value_type: 'Type', obj: 'Any') \
            -> 'Any':
        return "{" + ", ".join(
            key + ": " + value
            for key, value in context.convert_gen_map(
                key_type, value_type, obj, mapper=FLAT).items()) + "}"


class ArrayStringMapper(StringMapperBase):

    def data_record(
            self,
            context: 'Context',
            dt: 'DefDataType',
            record: 'DefDataType.Fields',
            obj: 'Any') -> 'Any':
        """
        Return the fields of this record as an array of strings.
        """
        elems = []
        for field in record.fields:
            val = context.convert(field.type, obj[field.field])
            if isinstance(val, str):
                elems.append(val)
            else:
                elems.extend(val)
        return elems


class FlatStringMapper(StringMapperBase):

    def data_record(
            self,
            context: 'Context',
            dt: 'DefDataType',
            record: 'DefDataType.Fields',
            obj: 'Any') -> 'Any':
        if record.fields:
            return "... with " + "; ".join(
                f"{field.field}: {context.convert(field.type, obj[field.field])}"
                for field in record.fields)
        else:
            return "..."


FLAT = FlatStringMapper()



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
The :mod:`dazl.util.prim_types` module contains functions for converting arbitrary objects and
on-wire formats to/from canonical Python types.
"""
import warnings
from decimal import Decimal

from datetime import datetime, date, timedelta
from typing import overload, Any, Mapping, Tuple, Union
from ..prim.datetime import TimeDeltaLike as TimeDeltaConvertible, DATETIME_ISO8601_Z_FORMAT

warnings.warn(
    'dazl.util.prim_types is deprecated; use dazl.prim instead.', DeprecationWarning, stacklevel=2)

__all__ = [
    'to_boolean', 'to_str', 'to_int', 'to_date', 'to_datetime', 'to_decimal', 'to_ledger_api_decimal',
    'to_timedelta', 'PrimitiveTypeConverter', 'decode_variant_dict', 'standard_format', 'nano_format',
    'unflatten_dotted_keys', 'DEFAULT_TYPE_CONVERTER', 'DATETIME_ISO8601_Z_FORMAT',
    'frozendict', 'to_hashable'
]


@overload
def to_boolean(obj: 'Union[bool, str]') -> bool: ...


@overload
def to_boolean(obj: None) -> None: ...


def to_boolean(obj):
    from ..prim import to_bool
    warnings.warn(
        'dazl.util.prim_types.to_boolean is deprecated; use dazl.prim.to_bool instead.',
        DeprecationWarning, stacklevel=2)
    return to_bool(obj) if obj is not None else None


def to_str(obj: Any) -> str:
    from ..prim import to_str
    warnings.warn(
        'dazl.util.prim_types.to_str is deprecated; use dazl.prim.to_str instead.',
        DeprecationWarning, stacklevel=2)
    return to_str(obj) if obj is not None else None


def to_int(obj: Any) -> int:
    from ..prim import to_int
    warnings.warn(
        "dazl.util.prim_types.to_int is deprecated; use dazl.prim.to_int instead.",
        DeprecationWarning, stacklevel=2)
    return to_int(obj) if obj is not None else None


def to_date(obj: Any) -> date:
    from ..prim import to_date
    warnings.warn(
        "dazl.util.prim_types.to_date is deprecated; use dazl.prim.to_date instead.",
        DeprecationWarning, stacklevel=2)
    return to_date(obj) if obj is not None else None


def to_datetime(obj: Any) -> datetime:
    from ..prim import to_datetime
    warnings.warn(
        "dazl.util.prim_types.to_datetime is deprecated; use dazl.prim.to_datetime instead.",
        DeprecationWarning, stacklevel=2)
    return to_datetime(obj) if obj is not None else None


def to_decimal(obj: Any) -> Decimal:
    from ..prim import to_decimal
    warnings.warn(
        "dazl.util.prim_types.to_decimal is deprecated; use dazl.prim.to_decimal instead.",
        DeprecationWarning, stacklevel=2)
    return to_decimal(obj) if obj is not None else None


def to_ledger_api_decimal(obj: 'Decimal') -> str:
    from ..prim import decimal_to_str
    warnings.warn(
        'dazl.util.prim_types.to_ledger_api_decimal is deprecated; use dazl.prim.decimal_to_str instead.',
        DeprecationWarning, stacklevel=2)
    return decimal_to_str(obj)


def to_timedelta(obj: TimeDeltaConvertible) -> 'timedelta':
    from ..prim import to_timedelta
    warnings.warn(
        'dazl.util.prim_types.to_timedelta is deprecated; use dazl.prim.to_timedelta instead.',
        DeprecationWarning, stacklevel=2)
    return to_timedelta(obj)


class PrimitiveTypeConverter:
    to_boolean = staticmethod(to_boolean)
    to_str = staticmethod(to_str)
    to_int = staticmethod(to_int)
    to_decimal = staticmethod(to_decimal)
    to_date = staticmethod(to_date)
    to_datetime = staticmethod(to_datetime)
    to_timedelta = staticmethod(to_timedelta)


def decode_variant_dict(obj: 'Any') -> 'Tuple[str, Any]':
    from ..prim import to_variant
    warnings.warn(
        'dazl.util.prim_types.decode_variant_dict is deprecated; use dazl.prim.to_variant instead.',
        DeprecationWarning, stacklevel=2)
    return to_variant(obj)


def standard_format(fmt, value):
    from ..prim.datetime import _parse
    warnings.warn(
        'dazl.util.prim_types.nano_format is deprecated; there is no replacement.',
        DeprecationWarning, stacklevel=2)
    return _parse(fmt, value)


def nano_format(value):
    from ..prim.datetime import _parse_nano_format
    warnings.warn(
        'dazl.util.prim_types.nano_format is deprecated; there is no replacement.',
        DeprecationWarning, stacklevel=2)
    return _parse_nano_format(value)


def unflatten_dotted_keys(m: 'Mapping[str, Any]'):
    from ..prim import to_record
    warnings.warn(
        'dazl.util.prim_types.unflatten_dotted_keys is deprecated; use dazl.prim.to_record instead.',
        DeprecationWarning, stacklevel=2)
    return to_record(m) if m is not None else None


DEFAULT_TYPE_CONVERTER = PrimitiveTypeConverter()


def frozendict(*args, **kwargs):
    from ..prim.map import FrozenDict
    warnings.warn(
        "dazl.util.prim_types.frozendict is deprecated; use dazl.prim.map.FrozenDict instead.",
        DeprecationWarning, stacklevel=2)
    return FrozenDict(*args, **kwargs)


def to_hashable(*args, **kwargs):
    from ..prim.map import to_hashable
    warnings.warn(
        "dazl.util.prim_types.to_hashable is deprecated; use dazl.prim.map.to_hashable instead.",
        DeprecationWarning, stacklevel=2)
    return to_hashable(*args, **kwargs)



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import Any, TYPE_CHECKING

from ..damlast.daml_lf_1 import DefDataType, Type

if TYPE_CHECKING:
    from .context import Context

if sys.version_info >= (3, 7):
    from typing import Protocol
else:
    from typing_extensions import Protocol

__all__ = ['ValueMapper']


class ValueMapper(Protocol):
    """
    Protocol for an object that can map one representation of DAML-LF values to another
    representation. This is the protocol that wire encoders/decoders adhere to, as well as certain
    types of validators.

    To use an instance of a :class:`ValueMapper` to transform an object, create a :class:`Context`
    with an instance of a :class:`ValueMapper`; then call :meth:`Context.convert`.
    """

    def data_record(
            self,
            context: 'Context',
            dt: 'DefDataType',
            record: 'DefDataType.Fields',
            obj: 'Any') \
            -> 'Any':
        raise NotImplementedError('data_record requires an implementation')

    def data_variant(
            self,
            context: 'Context',
            dt: 'DefDataType',
            variant: 'DefDataType.Fields',
            obj: 'Any') \
            -> 'Any':
        raise NotImplementedError('data_variant requires an implementation')

    def data_enum(self, context: 'Context', dt: 'DefDataType', enum: 'DefDataType.EnumConstructors',
                  obj: 'Any') -> 'Any':
        raise NotImplementedError('data_variant requires an implementation')

    def prim_unit(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_unit requires an implementation')

    def prim_bool(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_bool requires an implementation')

    def prim_int64(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_int64 requires an implementation')

    def prim_text(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_text requires an implementation')

    def prim_timestamp(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_timestamp requires an implementation')

    def prim_party(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_party requires an implementation')

    def prim_list(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_list requires an implementation')

    def prim_date(self, context: 'Context', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_date requires an implementation')

    def prim_contract_id(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_contract_id requires an implementation')

    def prim_optional(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_optional requires an implementation')

    def prim_text_map(self, context: 'Context', item_type: 'Type', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_text_map requires an implementation')

    def prim_numeric(self, context: 'Context', nat: int, obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_numeric requires an implementation')

    def prim_gen_map(
            self, context: 'Context', key_type: 'Type', value_type: 'Type', obj: 'Any') -> 'Any':
        raise NotImplementedError('prim_gen_map requires an implementation')



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Simple methods for instantiating an application written against dazl that incorporate common usage
patterns.
"""
import warnings
from argparse import ArgumentParser
from typing import Callable

from ..client.api import Network
from ..client.config import configure_parser, NetworkConfig


def run(init: 'Callable[[Network], None]'):
    warnings.warn("dazl.run is deprecated", DeprecationWarning, stacklevel=2)

    if init is None:
        raise ValueError('The init callback cannot be None')

    arg_parser = ArgumentParser()

    configure_parser(arg_parser, config_file_support=True)
    args = arg_parser.parse_args()

    config = NetworkConfig.get_config(args)
    network = Network()
    network.set_config(config)

    init(network)

    network.run_forever()



# Copyright (c) 2017-2020 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Contains primitive declarations and functions for working with "native" Python types as they
correspond to types over the Ledger API.
"""

from .basic import to_bool, to_str
from .complex import to_record, to_variant
from .contracts import ContractId, ContractData
from .datetime import TimeDeltaLike, date_to_int, date_to_str, datetime_to_str, \
    datetime_to_epoch_timedelta, datetime_to_epoch_microseconds, datetime_to_timestamp, \
    timedelta_to_duration, to_date, to_datetime, to_timedelta
from .json import JSONEncoder
from .map import FrozenDict
from .numbers import to_decimal, decimal_to_str, to_int
from .party import Party, to_party
